import base64, typing, re
from collections import OrderedDict

from ...nintendowifi import soapenvelopebase
from .manager import CtrSoapManager, ServiceLevel
from ..exception import ClassInitError, DataProcessingError
from .exception import OperationError, CTRExceptionBase
from ..certificate import Certificate
from ..ticket import Ticket
from ._common_parsers import _parse_attribute_member
from .._py_ver_fixes import LimitKind_T

#TODO:
# - GetTaxes
# - GetPurchaseInfo
# - GetTaxLocation
# - AccountListPurchaseHistory
# - AccountPurchaseTitle
# - AccountReplenishment

__all__ = [
	"AccountListETicketIds",
	"AccountGetETicketDetails",
	"AccountGetETickets",
	"AccountCheckBalance",
	"CheckECard",
	"CurrencyAccountsCheckBalance",
	"AccountDeleteTitleETickets",
	"DeleteSavedCard",
	"DownloadExpressETicket",
	"GetAccountStatus",
	"GetStandbyMode",
	"GetWalletInfo",
	"ETicketInfo",
	"SavedCardInfo"
]

class ETicketInfo:
	def __init__(
		self,
		ticketid: int,
		titleid: int,
		version: int,
		formatversion: int,
		migratecount: int,
		migratelimit: int,
		estimatedsize: int
	):
		self._ticketid = ticketid
		self._titleid = titleid
		self._version = version
		self._formatversion = formatversion
		self._migratecount = migratecount
		self._migratelimit = migratelimit
		self._estimatedsize = estimatedsize

	@property
	def __dict__(self):
		ret = OrderedDict()
		ret['TicketId'] = str(self._ticketid)
		ret['TitleId'] = f"{self._titleid:016X}"
		ret['Version'] = str(self._version)
		ret['FormatVersion'] = str(self._formatversion)
		ret['MigrateCount'] = str(self._migratecount)
		ret['MigrateLimit'] = str(self._migratelimit)
		ret['EstimatedSize'] = str(self._estimatedsize)
		return ret

	@property
	def ticketid(self) -> int:
		return self._ticketid

	@property
	def titleid(self) -> int:
		return self._titleid

	@property
	def version(self) -> int:
		return self._version

	@property
	def formatversion(self) -> int:
		return self._formatversion

	@property
	def migratecount(self) -> int:
		return self._migratecount

	@property
	def migratelimit(self) -> int:
		return self._migratelimit

	@property
	def estimatedsize(self) -> int:
		return self._estimatedsize

class SavedCardInfo:
	def __init__(
		self,
		cardtype: str,
		expirationmonth: str,
		expirationyear: str,
		maskedcardnumber: str
	):
		if len(cardtype) > 1 or len(expirationmonth) > 2 or \
			len(expirationyear) > 2 or len(maskedcardnumber) < 4:
			raise DataProcessingError("Invalid saved card information!")

		self._cardtype = cardtype
		self._expirationmonth = expirationmonth
		self._expirationyear = expirationyear
		self._maskedcardnumber = maskedcardnumber[-4:]

	@property
	def cardtype(self) -> str:
		return self._cardtype

	@property
	def expirationmonth(self) -> str:
		return self._expirationmonth

	@property
	def expirationyear(self) -> str:
		return self._expirationyear

	@property
	def maskedcardnumber(self) -> str:
		return self._maskedcardnumber

class TransactionInfo:
	def __init__(
		self,
		transactionid: int,
		date: int,
		_type: str
	):
		if not -0x8000000000000000 <= transactionid <= 0x7fffffffffffffff or \
			not -0x8000000000000000 <= date <= 0x7fffffffffffffff or \
			len(_type) > 15:
			raise DataProcessingError("Invalid transaction information!")

		self._transactionid = transactionid
		self._date = date
		self._type = _type

	@property
	def transactionid(self) -> int:
		return self._transactionid

	@property
	def date(self) -> int:
		return self._date

	@property
	def type(self) -> str:
		return self._type

def _xml_get_tiv_element(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> typing.Tuple[int,int]:
	text = parent._xml_get_str_element(parent, element)
	if not re.fullmatch("^[0-9]+\\.[0-9]+$", text):
		raise soapenvelopebase.XMLParseError("Invalid TIV found")

	split = text.split('.')
	return (int(split[0]), int(split[1]))

def _xml_get_eticketinfo_element(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> ETicketInfo:
	parent._xml_raise_if_text(element)
	ticketid = parent._xml_element_parse(element, 'urn:TicketId', parent._xml_get_int_element)
	titleid = parent._xml_element_parse(element, 'urn:TitleId', parent._xml_get_int_base16_element)
	version = parent._xml_element_parse(element, 'urn:Version', parent._xml_get_int_element)
	formatversion = parent._xml_element_parse(element, 'urn:FormatVersion', parent._xml_get_int_element, True)
	formatversion = formatversion if formatversion is not None else 0
	migratecount = parent._xml_element_parse(element, 'urn:MigrateCount', parent._xml_get_int_element)
	migratelimit = parent._xml_element_parse(element, 'urn:MigrateLimit', parent._xml_get_int_element)
	estimatedsize = parent._xml_element_parse(element, 'urn:EstimatedSize', parent._xml_get_int_element, True)
	estimatedsize = estimatedsize if estimatedsize is not None else 0
	return ETicketInfo(ticketid, titleid, version, formatversion, migratecount, migratelimit, estimatedsize)

# not to pass to a _xml_xxx_parse directly
def _xml_get_balance_element(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element,
	optional: bool
) -> typing.Optional[typing.Tuple[str,str]]:
	parent._xml_raise_if_text(element)
	amount = parent._xml_element_parse(element, 'urn:Amount', parent._xml_get_str_element, optional)
	currency = parent._xml_element_parse(element, 'urn:Currency', parent._xml_get_str_element, optional)
	if amount is None or currency is None:
		return None
	return (amount, currency)

def _xml_get_saved_card_element(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> SavedCardInfo:
	parent._xml_raise_if_text(element)
	cardtype = parent._xml_element_parse(element, 'urn:CardType', parent._xml_get_str_element)
	expirationmonth = parent._xml_element_parse(element, 'urn:ExpirationMonthMM', parent._xml_get_str_element)
	expirationyear = parent._xml_element_parse(element, 'urn:ExpirationYearYY', parent._xml_get_str_element)
	maskedcardnumber = parent._xml_element_parse(element, 'urn:MaskedCardNumber', parent._xml_get_str_element)
	return SavedCardInfo(cardtype, expirationmonth, expirationyear, maskedcardnumber)

# not to pass to a _xml_xxx_parse
def _parse_service_uris(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element, # that has 'ServiceURLs' in it
	optional: bool
) -> typing.Dict[str,str]:
	uris = {}
	def _parse_uri(
		parent: soapenvelopebase.SoapEnvelopeBase,
		element: soapenvelopebase.XML_Element
	) -> None:
		parent._xml_raise_if_text(element)
		name = parent._xml_element_parse(element, 'urn:Name', parent._xml_get_str_element)
		uri = parent._xml_element_parse(element, 'urn:URI', parent._xml_get_str_element)
		_name = {
			'EcsURL': 'ecs',
			'IasURL': 'ias',
			'CasURL': 'cas',
			'NusURL': 'nus',
			'ContentPrefixURL': 'content_prefix',
			'UncachedContentPrefixURL': 'uncached_content_prefix',
			'SystemContentPrefixURL': 'system_content_prefix',
			'SystemUncachedContentPrefixURL': 'system_uncached_content_prefix'
		}.get(name, None)
		if not _name:
			return
		uris[_name] = uri
	parent._xml_multi_element_parse_ret_none(element, 'urn:ServiceURLs', _parse_uri, optional)
	return uris

# not to pass to a _xml_xxx_parse
def _parse_account_attributes(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element, # that has 'AccountAttributes' in it
	optional: bool
) -> typing.Dict[str,typing.Optional[str]]:
	attributes = {}
	def _parse_attribute(
		parent: soapenvelopebase.SoapEnvelopeBase,
		element: soapenvelopebase.XML_Element
	) -> None:
		parent._xml_raise_if_text(element)
		name = parent._xml_element_parse(element, 'urn:Name', parent._xml_get_str_element)
		value = parent._xml_element_parse(element, 'urn:Value', parent._xml_get_str_element, True)
		_name = {
			'LOYALTY_LOGIN_NAME': 'LOYALTY_LOGIN_NAME'
		}.get(name, None)
		if not _name:
			return
		attributes[_name] = value
	parent._xml_multi_element_parse_ret_none(element, 'urn:AccountAttributes', _parse_attribute, optional)
	return attributes

def _xml_get_transaction_element(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> TransactionInfo:
	parent._xml_raise_if_text(element)
	transactionid = parent._xml_element_parse(element, 'urn:TransactionId', parent._xml_get_int_element)
	date = parent._xml_element_parse(element, 'urn:Date', parent._xml_get_int_element)
	_type = parent._xml_element_parse(element, 'urn:Type', parent._xml_get_str_element)
	return TransactionInfo(transactionid, date, _type)

def _xml_get_titleid_ticketid_pair(
	parent: soapenvelopebase.SoapEnvelopeBase,
	element: soapenvelopebase.XML_Element
) -> typing.Tuple[int, int]:
	parent._xml_raise_if_text(element)
	# titleid get checked for, but never checked or stored in actual code
	titleid = parent._xml_element_parse(element, 'urn:TitleId', parent._xml_get_int_base16_element)
	ticketid = parent._xml_element_parse(element, 'urn:TicketId', parent._xml_get_int_element)
	return (titleid, ticketid)

class AccountListETicketIds(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'AccountListETicketIds', ctrsoapmanager, True, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._tivs = self._xml_multi_element_parse(response, 'urn:TIV', _xml_get_tiv_element, True)
			self._tivs = tuple(self._tivs) if self._tivs is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def tivs(self) -> typing.Iterator[typing.Tuple[int, int]]:
		return self._tivs

class AccountGetETicketDetails(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, etickets: typing.Iterator[int]):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'AccountGetETicketDetails', ctrsoapmanager, True, True)

		for i in etickets:
			self._write_tag('TicketIds', i)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._eticketinfos = self._xml_multi_element_parse(response, 'urn:ETicketInfos', _xml_get_eticketinfo_element, True)
			self._eticketinfos = tuple(self._eticketinfos) if self._eticketinfos is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def eticketinfos(self) -> typing.Iterator[ETicketInfo]:
		return self._eticketinfos

class AccountGetETickets(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, etickets_ids: typing.Iterator[int]):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'AccountGetETickets', ctrsoapmanager, True, True)

		count = 0

		for i in etickets_ids:
			self._write_tag('TicketId', i)
			count += 1

		self._write_tag('DeviceCert', base64.b64encode(ctrsoapmanager.ct_cert).decode('utf-8'))

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._etickets = self._xml_multi_element_parse(response, 'urn:ETickets', lambda x,y: Ticket(self._xml_get_base64_element(x, y), ctrsoapmanager.ct_cert_full), True)
			self._etickets = tuple(self._etickets) if self._etickets is not None else tuple()
			self._certs = self._xml_multi_element_parse(response, 'urn:Certs', lambda x,y: Certificate(self._xml_get_base64_element(x, y)), True)
			self._certs = tuple(self._certs) if self._certs is not None else tuple()
			# by norm, ticket ids requested must be == etickets received, but here i'll ignore and leave a check field
			self._got_all = count == len(self._etickets)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def etickets(self) -> typing.Iterator[Ticket]:
		return self._etickets

	@property
	def certs(self) -> typing.Iterator[Certificate]:
		return self._certs

	@property
	def got_all(self) -> bool:
		return self._got_all

class AccountCheckBalance(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'AccountCheckBalance', ctrsoapmanager, False, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._balance = self._xml_element_parse(response, 'urn:Balance', lambda x, y: _xml_get_balance_element(x, y, False))
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def balance(self) -> typing.Tuple[str,str]:
		return self._balance

class CheckECard(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(
		self,
		ctrsoapmanager: CtrSoapManager,
		ecardid: str,
		ecardattributes: typing.Iterator[str]
	):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'CheckECard', ctrsoapmanager, True, True)

		self._write_tag('ECardId', ecardid)
		for i in ecardattributes:
			self._write_tag('ECardAttributes', i)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		if self.errorcode in [14, 15, 903]:
			self._validate_errorcode(self.errorcode, self.errormessage)

		# apparently we still parse not one of those errors stated above
		# but we still keep error

		try:
			response = response_parse[1]
			self._ecardattributes = self._xml_multi_element_parse(response, 'urn:ECardAttributes', _parse_attribute_member, True)
			self._ecardattributes = tuple(self._ecardattributes) if self._ecardattributes is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def ecardattributes(self) -> typing.Iterator[typing.Tuple[str, str]]:
		return self._ecardattributes

class CurrencyAccountsCheckBalance(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'CurrencyAccountsCheckBalance', ctrsoapmanager, False, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._balances = self._xml_multi_element_parse(response, 'urn:Balance', lambda x, y: _xml_get_balance_element(x, y, False), True)
			self._balances = tuple(self._balances) if self._balances is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def balances(self) -> typing.Iterator[typing.Tuple[str,str]]:
		return self._balances

class AccountDeleteTitleETickets(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, titleid: int):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'AccountDeleteTitleETickets', ctrsoapmanager, True, True)

		# normally its just one title, but plural makes me wonder if it supports multiple at once
		self._write_tag('Titles', f'{titleid:016X}')

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class DeleteSavedCard(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'DeleteSavedCard', ctrsoapmanager, True, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class DownloadExpressETicket(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(
		self,
		ctrsoapmanager: CtrSoapManager,
		itemid: int,
		titleid: int,
		limits: typing.Optional[
			typing.Iterable[
				typing.Tuple[
					int, # limits
					LimitKind_T # limit kind
				]
			]
		] = None,
		purchasenotes: typing.Optional[str] = None
	):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if not -0x80000000 <= itemid <= 0x7fffffff:
			raise DataProcessingError("Invalid item id")

		limits = limits if limits else ((0, 'PR'),)

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'DownloadExpressETicket', ctrsoapmanager, True, False)

		self._write_tag('ItemId', f"{itemid}")
		self._write_tag('TitleId', f"{titleid:016X}")

		for i in limits:
			if not -0x80000000 <= i[0] <= 0x7fffffff:
				raise DataProcessingError("Invalid limits")
			self._push_tag('Limits')
			self._write_tag('Limits', f"{i[0]}")
			self._write_tag('LimitKind', i[1][:4])
			self._pop_tag()

		self._write_tag('DeviceCert', base64.b64encode(ctrsoapmanager.ct_cert).decode('utf-8'))

		if ctrsoapmanager.age is not None:
			self._write_tag('Age', ctrsoapmanager.age)

		if purchasenotes is not None:
			self._write_tag('PurchaseNotes', purchasenotes)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		if self.errorcode in [14, 15, 903]:
			self._validate_errorcode(self.errorcode, self.errormessage)

		try:
			response = response_parse[1]
			self._transactions = self._xml_multi_element_parse(response, 'urn:Transactions', _xml_get_transaction_element, True)
			self._transactions = tuple(self._transactions) if self._transactions is not None else tuple()
			# ETickets implies multiple, but this one SOAP parses for one, just in case, I'll do multi parse
			self._etickets = self._xml_multi_element_parse(response, 'urn:ETickets', lambda x,y: Ticket(self._xml_get_base64_element(x, y), ctrsoapmanager.ct_cert_full))
			self._etickets = tuple(self._etickets) if self._etickets is not None else tuple()
			self._certs = self._xml_multi_element_parse(response, 'urn:Certs', lambda x,y: Certificate(self._xml_get_base64_element(x, y)), True)
			self._certs = tuple(self._certs) if self._certs is not None else tuple()
			id_pair = self._xml_element_parse(response, 'urn:TitleIdTicketId', _xml_get_titleid_ticketid_pair, True)
			self._titleid = id_pair[0] if id_pair else None
			self._ticketid = id_pair[1] if id_pair else None
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def transactions(self) -> typing.Iterable[TransactionInfo]:
		return self._transactions

	@property
	def etickets(self) -> typing.Iterable[Ticket]:
		return self._etickets

	@property
	def eticket(self) -> Ticket:
		return self._etickets[0]

	@property
	def certs(self) -> typing.Iterable[Certificate]:
		return self._certs

	@property
	def titleid(self) -> typing.Optional[int]:
		return self._titleid

	@property
	def ticketid(self) -> typing.Optional[int]:
		return self._ticketid

class GetAccountStatus(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'GetAccountStatus', ctrsoapmanager, True, True)

		self._write_tag('ECVersion', ctrsoapmanager.ec_version)
		self._write_tag('Locale', f"{ctrsoapmanager.language}_{ctrsoapmanager.country}")
		self._write_tag('ServiceLevel', ctrsoapmanager.service_level.name)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		#some error codes are ignored or parsed different
		#they are best validated after parsing values
		#most error codes are left for the callee to determine what to do

		response_parse = self._initiate_response_parse()

		try:
			response = response_parse[1]
			self._accountid = self._xml_element_parse(response, 'urn:AccountId', self._xml_get_int_element, True) if ctrsoapmanager.service_level == ServiceLevel.SHOP else None
			self._accountstatus = self._xml_element_parse(response, 'urn:AccountStatus', self._xml_get_str_element, True)
			if ctrsoapmanager.service_level == ServiceLevel.SHOP:
				self._balance = self._xml_element_parse(response, 'urn:Balance', lambda x, y: _xml_get_balance_element(x, y, True), True)
				self._eulaversion = self._xml_element_parse(response, 'urn:EulaVersion', self._xml_get_int_element, True)
				self._latesteulaversion = self._xml_element_parse(response, 'urn:LatestEulaVersion', self._xml_get_int_element, True)
				self._country = self._xml_element_parse(response, 'urn:Country', self._xml_get_str_element, True)
				self._region = self._xml_element_parse(response, 'urn:Region', self._xml_get_str_element, True) # Not originally parsed
				self._accountattributes = _parse_account_attributes(self, response, True)
				self._tivs = self._xml_multi_element_parse(response, 'urn:TIV', _xml_get_tiv_element, True)
				self._tivs = tuple(self._tivs) if self._tivs is not None else None
			self._uris = _parse_service_uris(self, response, True)
			self._ivssyncflag = self._xml_element_parse(response, 'urn:IVSSyncFlag', self._xml_get_bool_element, True)
			if self._ivssyncflag is None:
				self._ivssyncflag = False
			self._countryattribits = self._xml_element_parse(response, 'urn:CountryAttribits', self._xml_get_int_element, True) # Not originally parsed
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

		if not self._accountstatus and not self.servicestandbymode:
			raise soapenvelopebase.SoapError("No account status and not in service standby!")

		if not self._is_ready(ctrsoapmanager.service_level == ServiceLevel.SHOP):
			raise soapenvelopebase.SoapError("Soap response did not have all necessary information for the active service level")

		# validate errorcode
		if self.errorcode == 901 and self._accountstatus != 'U':
			raise soapenvelopebase.SoapError("Invalid combination of ErrorCode and AccountStatus")

	def _is_ready(self, is_shop: bool) -> bool:
		if is_shop:
			if self._uris.get('ecs', None) and \
				self._uris.get('cas', None) and \
				self._uris.get('content_prefix', None) and \
				self._uris.get('uncached_content_prefix', None) and \
				(self._uris.get('ias', None) or self.servicestandbymode):
				return True
		else:
			if self._uris.get('nus', None) and \
				self._uris.get('system_content_prefix', None) and \
				self._uris.get('system_uncached_content_prefix', None):
				return True
		return False

	def update_soap_managers_uris(self) -> None:
		if not self._uris:
			return

		for x, y in self._uris.items():
			self._session_manager.set_url_by_identifier(x, y)

	def validate_errorcode(self) -> None:
		self._validate_errorcode(self.errorcode, self.errormessage)

	@property
	def accountid(self) -> typing.Optional[int]:
		return self._accountid

	@property
	def accountstatus(self) -> typing.Optional[str]:
		return self._accountstatus

	@property
	def balance(self) -> typing.Optional[typing.Tuple[str,str]]:
		return self._balance

	@property
	def eulaversion(self) -> typing.Optional[int]:
		return self._eulaversion

	@property
	def latesteulaversion(self) -> typing.Optional[int]:
		return self._latesteulaversion

	@property
	def country(self) -> typing.Optional[str]:
		return self._country

	@property
	def region(self) -> typing.Optional[str]:
		return self._region

	@property
	def accountattributes(self) -> typing.Optional[typing.Dict[str,typing.Optional[str]]]:
		return self._accountattributes

	@property
	def tivs(self) -> typing.Optional[typing.Iterator[typing.Tuple[int, int]]]:
		return self._tivs

	@property
	def uris(self) -> typing.Optional[typing.Dict[str, str]]:
		return self._uris.copy() if self._uris is not None else None

	@property
	def ivssyncflag(self) -> bool:
		return self._ivssyncflag

	@property
	def countryattribits(self) -> typing.Optional[int]:
		return self._countryattribits

class GetStandbyMode(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'GetStandbyMode', ctrsoapmanager, False, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

		# this SOAP gets ServiceStandbyMode and TimeStamp responses mainly, which base already does.

class GetWalletInfo(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.ECS, 'GetWalletInfo', ctrsoapmanager, True, False)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ecs'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

		try:
			response = response_parse[1]
			self._savedcardinfo = self._xml_element_parse(response, 'urn:SavedCreditCardInfo', _xml_get_saved_card_element, True)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def savedcardinfo(self) -> typing.Optional[SavedCardInfo]:
		return self._savedcardinfo



import pathlib, typing

from ._env import *

__all__ = ['find_in_config_dirs']

def find_in_config_dirs(filepath: str, create_missing: bool = False) -> typing.List[str]:
	filepath = str(filepath)

	paths = []

	for i in config_dirs:
		try:
			path = (pathlib.Path(i) / filepath).resolve()
			if not path.is_file():
				continue

			paths.append(str(path))

		except Exception:
			pass

	if not paths and create_missing:
		for i in config_dirs:
			try:
				path = (pathlib.Path(i) / filepath).resolve()
				if not path.exists():
					path.parent.mkdir(parents=True, exist_ok=True)
					path.touch()

				if not path.is_file():
					continue

				paths.append(str(path))
				break

			except Exception:
				pass

	return list(set(paths))



import base64, typing

from ...nintendowifi import soapenvelopebase
from .manager import CtrSoapManager
from ..exception import ClassInitError, DataProcessingError
from .exception import OperationError, CTRExceptionBase
from ..movablesed import MovableSed
from ..regionaldata import Country, CountryType
from ._common_parsers import _parse_attribute_member_prefixed

# no IAS commands left to add

__all__ = [
	"GetChallenge",
	"GetRegistrationInfo",
	"Register",
	"Unregister",
	"AccountTransfer",
	"MoveAccount",
	"SetIVSData",
	"GetIVSData",
	"ReportIVSSync",
	"SetCountry",
	"SetExternalAccount",
	"DeleteExternalAccount",
	"GetAccountAttributesByProfile"
]

class GetChallenge(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'GetChallenge', ctrsoapmanager, False, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._challenge = self._xml_element_parse(response, 'urn:Challenge', self._xml_get_int_element)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def challenge(self) -> int:
		return self._challenge

class GetRegistrationInfo(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, challenge: int):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if not isinstance(challenge, int):
			raise ClassInitError("Challenge value must be an int returned from GetChallenge soap, and not a reuse")

		signed_message_block = '<Challenge>{0}</Challenge>'.format(challenge)

		signature_data = ctrsoapmanager.ap_sign(0x0004013000002c02, signed_message_block.encode('utf-8'))

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'GetRegistrationInfo', ctrsoapmanager, True, True)

		self._write_tag('DeviceCert', base64.b64encode(ctrsoapmanager.ct_cert).decode('utf-8'))
		self._write_tag('Signature', base64.b64encode(signature_data[1]).decode('utf-8'))
		self._write_tag('CertChain', base64.b64encode(signature_data[0]).decode('utf-8'))
		self._write_tag('Challenge', challenge)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()

		if self.errorcode == 901: # 'IAS - Device Id is not registered'
			self._is_registered = False
			self._accountid = None
			self._devicetoken = None
			self._devicetokenexpired = None
			self._country = None
			self._extaccountid = None
			self._devicecode = None
			return
		else:
			self._is_registered = True # unless other error, which _validate_errorcode will catch

		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._accountid = self._xml_element_parse(response, 'urn:AccountId', self._xml_get_int_element)
			self._devicetoken = self._xml_element_parse(response, 'urn:DeviceToken', self._xml_get_str_element)
			self._devicetokenexpired = self._xml_element_parse(response, 'urn:DeviceTokenExpired', self._xml_get_bool_element, True)
			self._country = self._xml_element_parse(response, 'urn:Country', self._xml_get_str_element)
			self._extaccountid = self._xml_element_parse(response, 'urn:ExtAccountId', self._xml_get_str_element, True)
			self._devicecode = self._xml_element_parse(response, 'urn:DeviceCode', self._xml_get_str_element, True)
			if self._devicetokenexpired is None:
				self._devicetokenexpired = False
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def accountid(self) -> typing.Optional[int]:
		return self._accountid

	@property
	def devicetoken(self) -> typing.Optional[str]:
		return self._devicetoken

	@property
	def devicetokenexpired(self) -> bool:
		return self._devicetokenexpired

	@property
	def country(self) -> typing.Optional[str]:
		return self._country

	@property
	def extaccountid(self) -> typing.Optional[str]:
		return self._extaccountid

	@property
	def devicecode(self) -> typing.Optional[str]:
		return self._devicecode

	@property
	def is_registered(self) -> bool:
		return self._is_registered

class Register(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, challenge: int):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if not isinstance(challenge, int):
			raise ClassInitError("Challenge value must be an int returned from GetChallenge soap, and not a reuse")

		signed_message_block =  '<Challenge>{0}</Challenge>'.format(challenge)
		signed_message_block += '<SerialNumber>{0}</SerialNumber>'.format(ctrsoapmanager.serial_no)
		signed_message_block += '<Country>{0}</Country>'.format(ctrsoapmanager.country)
		signed_message_block += '<RegisterRegion>{0}</RegisterRegion>'.format(ctrsoapmanager.region)
		signed_message_block += '<Language>{0}</Language>'.format(ctrsoapmanager.language)

		signature_data = ctrsoapmanager.ap_sign(0x0004013000002c02, signed_message_block.encode('utf-8'))

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'Register', ctrsoapmanager, True, True)

		self._write_tag('DeviceCert', base64.b64encode(ctrsoapmanager.ct_cert).decode('utf-8'))
		self._write_tag('SerialNumber', ctrsoapmanager.serial_no)
		self._write_tag('RegisterRegion', ctrsoapmanager.region)
		self._write_tag('Signature', base64.b64encode(signature_data[1]).decode('utf-8'))
		self._write_tag('CertChain', base64.b64encode(signature_data[0]).decode('utf-8'))
		self._write_tag('Challenge', challenge)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._accountid = self._xml_element_parse(response, 'urn:AccountId', self._xml_get_int_element)
			self._devicetoken = self._xml_element_parse(response, 'urn:DeviceToken', self._xml_get_str_element)
			self._devicetokenexpired = self._xml_element_parse(response, 'urn:DeviceTokenExpired', self._xml_get_bool_element, True)
			self._country = self._xml_element_parse(response, 'urn:Country', self._xml_get_str_element)
			self._extaccountid = self._xml_element_parse(response, 'urn:ExtAccountId', self._xml_get_str_element, True)
			self._devicecode = self._xml_element_parse(response, 'urn:DeviceCode', self._xml_get_str_element, True)
			if self._devicetokenexpired is None:
				self._devicetokenexpired = False
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def accountid(self) -> int:
		return self._accountid

	@property
	def devicetoken(self) -> str:
		return self._devicetoken

	@property
	def devicetokenexpired(self) -> bool:
		return self._devicetokenexpired

	@property
	def country(self) -> str:
		return self._country

	@property
	def extaccountid(self) -> typing.Optional[str]:
		return self._extaccountid

	@property
	def devicecode(self) -> typing.Optional[str]:
		return self._devicecode

	@property
	def is_registered(self) -> bool:
		return True

class Unregister(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, challenge: int):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if not isinstance(challenge, int):
			raise ClassInitError("Challenge value must be an int returned from GetChallenge soap, and not a reuse")

		signed_message_block =  '<Challenge>{0}</Challenge>'.format(challenge)
		signed_message_block += '<SerialNumber>{0}</SerialNumber>'.format(ctrsoapmanager.serial_no)

		signature_data = ctrsoapmanager.ap_sign(0x0004013000002c02, signed_message_block.encode('utf-8'))

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'Unregister', ctrsoapmanager, True, True)

		self._write_tag('DeviceCert', base64.b64encode(ctrsoapmanager.ct_cert).decode('utf-8'))
		self._write_tag('SerialNumber', ctrsoapmanager.serial_no)
		self._write_tag('Signature', base64.b64encode(signature_data[1]).decode('utf-8'))
		self._write_tag('CertChain', base64.b64encode(signature_data[0]).decode('utf-8'))
		self._write_tag('Challenge', challenge)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class AccountTransfer(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, challenge: int):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if not isinstance(challenge, int):
			raise ClassInitError("Challenge value must be an int returned from GetChallenge soap, and not a reuse")

		signed_message_block =  '<Challenge>{0}</Challenge>'.format(challenge)
		signed_message_block += '<SerialNumber>{0}</SerialNumber>'.format(ctrsoapmanager.serial_no)
		signed_message_block += '<Country>{0}</Country>'.format(ctrsoapmanager.country)
		signed_message_block += '<RegisterRegion>{0}</RegisterRegion>'.format(ctrsoapmanager.region)
		signed_message_block += '<Language>{0}</Language>'.format(ctrsoapmanager.language)

		signature_data = ctrsoapmanager.ap_sign(0x0004013000002c02, signed_message_block.encode('utf-8'))

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'AccountTransfer', ctrsoapmanager, True, True)

		self._write_tag('DeviceCert', base64.b64encode(ctrsoapmanager.ct_cert).decode('utf-8'))
		self._write_tag('SerialNumber', ctrsoapmanager.serial_no)
		self._write_tag('RegisterRegion', ctrsoapmanager.region)
		self._write_tag('Signature', base64.b64encode(signature_data[1]).decode('utf-8'))
		self._write_tag('CertChain', base64.b64encode(signature_data[0]).decode('utf-8'))
		self._write_tag('Challenge', challenge)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._accountid = self._xml_element_parse(response, 'urn:AccountId', self._xml_get_int_element)
			self._devicetoken = self._xml_element_parse(response, 'urn:DeviceToken', self._xml_get_str_element)
			self._devicetokenexpired = self._xml_element_parse(response, 'urn:DeviceTokenExpired', self._xml_get_bool_element, True)
			self._country = self._xml_element_parse(response, 'urn:Country', self._xml_get_str_element)
			self._extaccountid = self._xml_element_parse(response, 'urn:ExtAccountId', self._xml_get_str_element, True)
			self._devicecode = self._xml_element_parse(response, 'urn:DeviceCode', self._xml_get_str_element, True)
			if self._devicetokenexpired is None:
				self._devicetokenexpired = False
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def accountid(self) -> int:
		return self._accountid

	@property
	def devicetoken(self) -> str:
		return self._devicetoken

	@property
	def devicetokenexpired(self) -> bool:
		return self._devicetokenexpired

	@property
	def country(self) -> str:
		return self._country

	@property
	def extaccountid(self) -> typing.Optional[str]:
		return self._extaccountid

	@property
	def devicecode(self) -> typing.Optional[str]:
		return self._devicecode

class MoveAccount(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(
		self,
		ctrsoapmanager: CtrSoapManager,
		target_device_id: int,
		target_account_id: int,
		target_device_token: str,
		check_only: bool
	):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'MoveAccount', ctrsoapmanager, True, True)

		self._write_tag('TargetDeviceId', str(int(target_device_id)))
		self._write_tag('TargetAccountId', str(int(target_account_id)))
		self._write_tag('TargetDeviceToken', str(target_device_token))
		if check_only:
			self._write_tag('CheckOnly', 'true')

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		# on something like NIM, errorcode is parsed as type of reason of why we can or cannot transfer
		# for now, raise error
		self._validate_errorcode(self.errorcode, self.errormessage)

class SetIVSData(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, ivs: typing.SupportsBytes):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'SetIVSData', ctrsoapmanager, True, True)

		self._write_tag('IVSData', base64.b64encode(bytes(ivs)).decode('utf-8'))

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class GetIVSData(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'GetIVSData', ctrsoapmanager, True, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			# originally, this is considered "optional", error still rises, not really "optional"
			self._ivsdata = self._xml_element_parse(response, 'urn:IVSData', self._xml_get_base64_element, True)
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def ivsdata(self) -> typing.Optional[bytes]:
		return self._ivsdata

class ReportIVSSync(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'ReportIVSSync', ctrsoapmanager, True, True)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class SetCountry(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, countrycode: CountryType):
		countrycode = Country.get_country_str(countrycode)

		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		if countrycode is None: # NIM accepts any string, however, I do not
			raise ClassInitError("Invalid country code")

		_overrides = ctrsoapmanager.get_overrides_copy()

		try:
			ctrsoapmanager.set_country_override(countrycode)

			super().__init__(soapenvelopebase.SoapSubNames.IAS, 'SetCountry', ctrsoapmanager, True, True)

			self._write_tag('CountryCode', countrycode)

			ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
			if ret != 200:
				raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

			response_parse = self._initiate_response_parse()
			self._validate_errorcode(self.errorcode, self.errormessage)

			ctrsoapmanager.commit_overrides()
		except:
			ctrsoapmanager.load_overrides_copy(_overrides)
			raise

class SetExternalAccount(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, ext_account_id: str, password: str):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'SetExternalAccount', ctrsoapmanager, True, True)

		self._write_tag('ExtAccountId', ext_account_id)
		self._write_tag('NewPassword', password)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class DeleteExternalAccount(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, ext_account_id: str):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'DeleteExternalAccount', ctrsoapmanager, True, False)

		self._write_tag('ExtAccountId', ext_account_id)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)

class GetAccountAttributesByProfile(soapenvelopebase.SoapEnvelopeBase):
	@soapenvelopebase.ObjectTimingEmuHelper(0.9, 0.125)
	def __init__(self, ctrsoapmanager: CtrSoapManager, profile: str):
		if not isinstance(ctrsoapmanager, CtrSoapManager):
			raise ClassInitError("Expected CtrSoapManager")

		super().__init__(soapenvelopebase.SoapSubNames.IAS, 'GetAccountAttributesByProfile', ctrsoapmanager, False, False)

		self._write_tag('Profile', profile)

		ret = self._send(ctrsoapmanager.get_url_by_identifier('ias'))
		if ret != 200:
			raise OperationError("Bad HTTP response or connection error, ret = {0}".format(ret))

		response_parse = self._initiate_response_parse()
		self._validate_errorcode(self.errorcode, self.errormessage)
		try:
			response = response_parse[1]
			self._accountattribits = self._xml_element_parse(response, 'urn:AccountAttribits', self._xml_get_int_element, True)
			self._accountattribits = 0 if self._accountattribits is None else self._accountattribits
			if self._accountattribits < 0 or self._accountattribits > 0xffffffffffffffff:
				raise DataProcessingError("Out of range account attribits")
			self._accountattributes = self._xml_multi_element_parse(response, 'urn:AccountAttributes', _parse_attribute_member_prefixed, True)
			self._accountattributes = tuple(self._accountattributes) if self._accountattributes is not None else tuple()
		except CTRExceptionBase:
			raise
		except Exception as e:
			raise soapenvelopebase.XMLParseError("Unexpected exception while parsing XML") from e

	@property
	def accountattribits(self) -> int:
		return self._accountattribits

	@property
	def accountattributes(self) -> typing.Iterator[typing.Tuple[str, str]]:
		return self._accountattributes



from struct import pack, unpack
import typing, hashlib

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives import hashes, cmac
import cryptography.hazmat.primitives.ciphers as crypto_ciphers

from . import keys
from .lfcs import LFCS
from .otp import OTP
from .exception import ClassInitError, DataProcessingError
from .constants import _load_p9_keys, _p9_keys

__all__ = [
	"MovableSed"
]

_load_p9_keys()

class MovableSed:
	def __init__(self, data: typing.SupportsBytes):
		self._msed_data = bytes(data)[:0x140]

		if 0x120 != len(self._msed_data) != 0x140:
			raise ClassInitError("movable.sed should be 288 or 320 bytes")

		segments = unpack("<4s4B272sQ", self._msed_data[:0x120])

		if segments[0] != b'SEED':
			raise ClassInitError("movable.sed invalid start")

		if not (segments[1] == segments[3] == segments[4] == 0):
			# these should not have effects, but they still should be 0
			# asides not allowing non 0 if @ 0x5 is 0
			raise ClassInitError("movable.sed byte @ 0x4, 0x6 and 0x7 should be 0")

		if 0 != segments[2] != 1:
			raise ClassInitError("movable.sed unrecognized byte value @ 0x5, expected 0 or 1")

		self._lfcs = LFCS(segments[5])

		if segments[2] == 1 and len(self._msed_data) == 0x140:
			other_segments = unpack("<I12s16s", self._msed_data[0x120:])
			if sum(other_segments[1]) != 0:
				raise ClassInitError("movable.sed the 12 bytes @ 0x124 should be all 0")
			self._format_rng_increment = other_segments[0]
			self._cmac = other_segments[2]
		elif segments[2] == 1 and len(self._msed_data) == 0x120:
			segments = list(segments)
			segments[2] = 0
			segments = tuple(segments)
			self._msed_data = pack("<4s4B272sQ", *segments)
			self._format_rng_increment = None
			self._cmac = None
		else:
			self._format_rng_increment = None
			self._cmac = None

		self._high_u64 = segments[6]
		self._keyy = pack("<QQ", self._lfcs.lfcs_id, self._high_u64)

	def __bytes__(self):
		return self._msed_data

	def __len__(self):
		return len(self._msed_data)

	def __repr__(self):
		modname = self.__class__.__module__
		clsname = self.__class__.__qualname__
		return f"<{modname}.{clsname} lfcs={self._lfcs!r} high_u64={self._high_u64!r} keyy={self._keyy!r} format_rng_increment={self._format_rng_increment!r} cmac={self._cmac!r}>"

	@classmethod
	def generate_from_otp(cls, otp: OTP) -> 'MovableSed':
		if not isinstance(otp, OTP):
			raise DataProcessingError("Expected OTP")

		lfcs = LFCS.generate_from_otp(otp)

		msed = pack("<4s4x272sQ", b'SEED', bytes(lfcs), otp.supplementary_id)

		msed = cls(msed)
		msed.fixup_footer_data(otp)

		return msed

	@keys._ensure_b9(False)
	def fixup_footer_data(self, otp: OTP, dev: typing.Optional[bool] = None) -> bool:
		if not isinstance(otp, OTP):
			raise DataProcessingError("Expected OTP")

		if not otp.gen_console_keys():
			return False

		key = keys.scramble_keys(
			otp.console_keys_x[0xB],
			keys.get_b9_y(0xB, dev if dev is not None else otp.is_dev)
		)

		if key is None:
			return False

		key = key.to_bytes(16, 'big')

		format_rng_increment = 0
		if self._lfcs.lfcs_id == otp.lfcs_id:
			format_rng_increment = self._high_u64 - otp.supplementary_id
			format_rng_increment &= 0xFFFFFFFF

		msed = pack("<4sxB2x272sQI12x", b'SEED', 1, bytes(self._lfcs), self._high_u64, format_rng_increment)

		c = cmac.CMAC(crypto_ciphers.algorithms.AES(key), default_backend())
		c.update(hashlib.sha256(msed).digest())
		cmac_hash = c.finalize()

		msed += cmac_hash

		self._msed_data = msed
		self._format_rng_increment = format_rng_increment
		self._cmac = cmac_hash

		return True

	@keys._ensure_b9(False)
	def check_cmac(
		self,
		otp: OTP,
		dev: typing.Optional[bool] = None,
		fix: bool = False
	) -> typing.Optional[bool]:
		if self._cmac is None:
			if fix:
				return self.fixup_footer_data(otp, dev)
			return None

		if not isinstance(otp, OTP):
			raise DataProcessingError("Expected OTP")

		if not otp.gen_console_keys():
			return False

		key = keys.scramble_keys(
			otp.console_keys_x[0xB],
			keys.get_b9_y(0xB, dev if dev is not None else otp.is_dev)
		)

		if key is None:
			return False

		key = key.to_bytes(16, 'big')

		c = cmac.CMAC(crypto_ciphers.algorithms.AES(key), default_backend())
		c.update(hashlib.sha256(self._msed_data[:0x130]).digest())
		cmac_hash = c.finalize()

		if fix and self._cmac != cmac_hash:
			self._msed_data = self._msed_data[:0x130] + cmac_hash
			self._cmac = cmac_hash

		return self._cmac == cmac_hash

	@classmethod
	@keys._ensure_b9(None)
	def load_from_ivs(cls, ivs: typing.SupportsBytes, dev: bool) -> typing.Optional['MovableSed']:
		ivs = bytes(ivs)[:0x130]
		if len(ivs) != 0x130:
			return None

		keyx_0x35 = keys.get_b9_x(0x35, dev)

		keyy = _p9_keys['IVS_KEYY']
		cmackeyy = _p9_keys['IVS_CMAC_KEYY']
		if keyy is None or cmackeyy is None:
			_load_p9_keys()
			keyy = _p9_keys['IVS_KEYY']
			cmackeyy = _p9_keys['IVS_CMAC_KEYY']
			if keyy is None or cmackeyy is None:
				return None

		key = keys.scramble_keys(
			keyx_0x35,
			keyy
		)

		cmackey = keys.scramble_keys(
			keyx_0x35,
			cmackeyy
		)

		if key is None or cmackey is None:
			return None

		key = key.to_bytes(16, 'big')
		cmackey = cmackey.to_bytes(16, 'big')

		decryptor = crypto_ciphers.Cipher(
			crypto_ciphers.algorithms.AES(key),
			crypto_ciphers.modes.CBC(ivs[:0x10]),
			default_backend()
		).decryptor()
		msed = decryptor.update(ivs[0x10:]) + decryptor.finalize()

		c = cmac.CMAC(crypto_ciphers.algorithms.AES(cmackey), default_backend())
		c.update(hashlib.sha256(msed[:0x110]).digest())
		cmac_hash = c.finalize()

		if cmac_hash != ivs[:0x10]:
			return None

		return cls(msed)

	@keys._ensure_b9(None)
	def produce_ivs(self, dev: typing.Optional[bool] = None) -> typing.Optional[bytes]:
		keyx_0x35 = keys.get_b9_x(0x35, dev if dev is not None else self._lfcs.is_dev)

		keyy = _p9_keys['IVS_KEYY']
		cmackeyy = _p9_keys['IVS_CMAC_KEYY']
		if keyy is None or cmackeyy is None:
			_load_p9_keys()
			keyy = _p9_keys['IVS_KEYY']
			cmackeyy = _p9_keys['IVS_CMAC_KEYY']
			if keyy is None or cmackeyy is None:
				return None

		key = keys.scramble_keys(
			keyx_0x35,
			keyy
		)

		cmackey = keys.scramble_keys(
			keyx_0x35,
			cmackeyy
		)

		if key is None or cmackey is None:
			return None

		key = key.to_bytes(16, 'big')
		cmackey = cmackey.to_bytes(16, 'big')

		c = cmac.CMAC(crypto_ciphers.algorithms.AES(cmackey), default_backend())
		c.update(hashlib.sha256(self._msed_data[:0x110]).digest())
		cmac_hash = c.finalize()

		encryptor = crypto_ciphers.Cipher(
			crypto_ciphers.algorithms.AES(key),
			crypto_ciphers.modes.CBC(cmac_hash),
			default_backend()
		).encryptor()
		ivs_msed = encryptor.update(self._msed_data[:0x120]) + encryptor.finalize()

		return cmac_hash + ivs_msed

	@property
	def lfcs(self) -> LFCS:
		return self._lfcs

	@property
	def high_u64(self) -> int:
		return self._high_u64

	@property
	def keyy(self) -> bytes:
		return self._keyy

	@property
	def format_rng_increment(self) -> typing.Optional[int]:
		return self._format_rng_increment

	@property
	def cmac(self) -> typing.Optional[bytes]:
		return self._cmac



from enum import IntEnum, unique
import typing, os, base64, abc

from cryptography.hazmat.backends import default_backend
from cryptography.hazmat.primitives.asymmetric import rsa
import cryptography.hazmat.primitives.ciphers as crypto_ciphers

from .ticket import Ticket
from .title import Title
from .tmd import TMD, ContentChunkRecord
from .certificate import Certificate
from .exception import CTRExceptionBase, ClassInitError, DataProcessingError
from .ssl import _ssl_certs
from ..connection import Connection, SimpleDownloadBuffer, MultiWriter, WriterFuncType
from .constants import _load_p9_keys, _p9_pubnums

__all__ = [
	"CDN",
	"CDNDownloadStage",
	"AssistedCDNDownload"
]

_load_p9_keys()

def _CDNTicketEnc(tik):
	if (((len(tik) + 15) >> 4) << 4) >= 0x2800:
		raise DataProcessingError("3DS caps tickets to be sent by 0x2800")

	tik = bytes(tik)
	tik += bytes((16 - (len(tik) & 0xF)) & 0xF)
	key_iv = os.urandom(32)

	cdnkey = _p9_pubnums['CDN_TIK_WRAP']
	if cdnkey is None:
		_load_p9_keys()
		cdnkey = _p9_pubnums['CDN_TIK_WRAP']
		if cdnkey is None:
			raise DataProcessingError("No CDN Key!! Provide up-to-date P9 code.bin to load function")

	klen = (cdnkey.n.bit_length() + 7) // 8
	plen = klen - len(key_iv) - 3
	if plen <= 0:
		raise DataProcessingError("Bad Key? Was it modified?")

	m = b'\x00\x01' + b'\xFF' * plen + b'\x00' + key_iv
	m = int.from_bytes(m, 'big')
	m = pow(m, cdnkey.e, cdnkey.n).to_bytes(klen, 'big')

	encryptor = crypto_ciphers.Cipher(
		crypto_ciphers.algorithms.AES(key_iv[:16]),
		crypto_ciphers.modes.CBC(key_iv[16:]),
		default_backend()
	).encryptor()

	tik_enc = encryptor.update(tik) + encryptor.finalize()

	return (base64.b64encode(m).decode('ascii'), base64.b64encode(tik_enc).decode('ascii'))

@unique
class CDNDownloadStage(IntEnum):
	CETK = 0
	TMD = 1
	CONTENT = 2
	END = 3

class AssistedCDNDownload(metaclass=abc.ABCMeta):
	@abc.abstractmethod
	def notify(
		self,
		stage: CDNDownloadStage,
		chunk: ContentChunkRecord,
		starting: bool,
		http_ret: typing.SupportsInt,
		cdn: 'CDN'
	) -> typing.Optional[WriterFuncType]:
		"""
		Notify the stage
		chunk record to be downloaded if any
		if starting or ending
		if ending, what http return we got
		reference to CDN object
		return writer if starting and if applicable
		"""

	@abc.abstractmethod
	def in_error(self) -> bool:
		"""
		Check if object is in error state
		"""

	@abc.abstractmethod
	def abort(self) -> None:
		"""
		Raise abort request
		"""

class CDN:
	def __init__(
		self,
		title_id: typing.SupportsInt,
		version: typing.Optional[typing.SupportsInt] = None,
		*,
		device_id: typing.Optional[typing.SupportsInt] = None,
		account_id: typing.Optional[typing.SupportsInt] = None,
		ticket: typing.Optional[Ticket] = None,
		content_prefix: typing.Optional[str] = None,
		uncached_content_prefix: typing.Optional[str] = None,
		system_content_prefix: typing.Optional[str] = None,
		system_uncached_content_prefix: typing.Optional[str] = None,
		is_dev: bool = False
	):
		self._tid = int(title_id) & 0xFFFFFFFFFFFFFFFF

		if not Title.is_ctr(self._tid):
			raise ClassInitError("Expected ctr title")

		self._version = (int(version) & 0xFFFF) if version is not None else None

		self._device_id = ((4<<32) | (int(device_id) & 0xFFFFFFFF)) if device_id else None
		self._account_id = (int(account_id) & 0xFFFFFFFF) if account_id else None

		if ticket is not None and not isinstance(ticket, Ticket):
			raise ClassInitError("ticket was given but is not a class Ticket")

		if ticket.title_id != self._tid:
			raise ClassInitError("ticket's title id does not match CDN download target")

		self._tik = ticket

		self._urls = {
			'cdn': str(content_prefix) if content_prefix is not None else 'http://ccs.cdn.c.shop.nintendowifi.net/ccs/download',
			'uncached_cdn': str(uncached_content_prefix) if uncached_content_prefix is not None else 'https://ccs.c.shop.nintendowifi.net/ccs/download',
			'syscdn': str(system_content_prefix) if system_content_prefix is not None else 'http://nus.cdn.c.shop.nintendowifi.net/ccs/download',
			'uncached_syscdn': str(system_uncached_content_prefix) if system_uncached_content_prefix is not None else 'https://ccs.c.shop.nintendowifi.net/ccs/download'
		}

		self._tmd_raw = None
		self._tmd = None
		self._cetk_raw = None
		self._certs = {}

		self._is_dev = bool(is_dev)

		self._conn = self._create_connection(self._is_dev)

	@staticmethod
	def _create_connection(is_dev: bool = False) -> Connection:
		conn = Connection()
		conn.set_keepalive(True)
		conn.set_cainfo(_ssl_certs._ca_id_path(3))
		conn.set_cli_cert(*_ssl_certs._client_cert_path_tuple(is_dev))

		return conn

	def _update_certchain(self, chainblock: typing.SupportsBytes) -> None:
		certs = Certificate.process_chain(chainblock)
		for i in certs:
			self.add_additional_cert(i)

	@property
	def target_version(self):
		return self._version

	@property
	def tmd(self) -> typing.Optional[TMD]:
		return self._tmd

	@property
	def ticket(self) -> typing.Optional[Ticket]:
		return self._tik

	@property
	def tmd_raw(self) -> typing.Optional[bytes]:
		return self._tmd_raw

	@property
	def ticket_raw(self) -> typing.Optional[bytes]:
		return self._cetk_raw

	@property
	def certs(self) -> typing.Dict[str, Certificate]:
		return self._certs.copy()

	@property
	def is_sys_tid(self) -> bool:
		return Title.is_ctranysys(self._tid)

	@property
	def ignore_tik(self) -> bool:
		return Title.is_ctranysys(self._tid) or Title.is_ctrdlp(self._tid)

	@property
	def tmd_url(self) -> str:
		base_url = self._urls['uncached_syscdn' if self.is_sys_tid else 'uncached_cdn']
		url = base_url + f'/{self._tid:016X}/tmd'
		if not self._device_id and not self._account_id:
			return url
		url += '?'
		if self._device_id:
			url += f'deviceId={self._device_id}'
		if self._device_id and self._account_id:
			url += '&'
		if self._account_id:
			url += f'accountId={self._account_id}'
		return url

	@property
	def cetk_url(self) -> typing.Optional[str]:
		if not self.is_sys_tid:
			return None

		return self._urls['syscdn'] + f'/{self._tid:016X}/cetk'

	def content_url(self, c_id: typing.SupportsInt) -> str:
		base_url = self._urls['syscdn' if self.is_sys_tid else 'cdn']
		return base_url + f'/{self._tid:016X}/{(int(c_id) & 0xFFFFFFFF):08X}'

	def download_tmd(self, write: typing.Optional[WriterFuncType] = None) -> int:
		if write is not None and not callable(write):
			raise DataProcessingError("non callable writer")

		buf = SimpleDownloadBuffer()
		write_func = buf.write if write is None else MultiWriter([buf.write, write]).write

		self._conn.reset_headers()
		self._conn.set_url(self.tmd_url)
		self._conn.set_write_function(write_func)
		ret = self._conn.perform()

		if ret != 200:
			return ret

		self._tmd_raw = buf.get()
		self._tmd = TMD(self._tmd_raw)
		self._update_certchain(self._tmd_raw[len(self._tmd):])

		return ret

	def download_cetk(self, write: typing.Optional[WriterFuncType] = None) -> typing.Optional[int]:
		if write is not None and not callable(write):
			raise DataProcessingError("non callable writer")

		url = self.cetk_url
		if url is None:
			return None

		buf = SimpleDownloadBuffer()
		write_func = buf.write if write is None else MultiWriter([buf.write, write]).write

		self._conn.reset_headers()
		self._conn.set_url(url)
		self._conn.set_write_function(write_func)
		ret = self._conn.perform()

		if ret != 200:
			return ret

		self._cetk_raw = buf.get()
		self._tik = Ticket(self._cetk_raw)
		self._update_certchain(self._cetk_raw[len(self._tik):])

		return ret

	def download_content_id(self, c_id: typing.SupportsInt, write_func: typing.Optional[WriterFuncType]) -> int:
		if write_func is not None and not callable(write_func):
			raise DataProcessingError("non callable writer")

		self._conn.reset_headers()
		self._conn.set_url(self.content_url(c_id))
		if not self.ignore_tik and self._tik:
			b64_enckey, b64_enctik = _CDNTicketEnc(self._tik)
			self._conn.set_header("X-Authentication-Key", b64_enckey)
			self._conn.set_header("X-Authentication-Data", b64_enctik)
		self._conn.set_nobody(write_func is None)
		self._conn.set_write_function((lambda x: None) if write_func is None else write_func)
		ret = self._conn.perform()

		return ret

	def download(
		self,
		assistance: AssistedCDNDownload,
		right_check: bool = True,
		content_error_continue: bool = False
	) -> bool:
		if not isinstance(assistance, AssistedCDNDownload):
			raise DataProcessingError("unexpected assistance type")

		if self.is_sys_tid:
			writer = assistance.notify(CDNDownloadStage.CETK, None, True, 0, self)
			if writer is None and assistance.in_error():
				assistance.abort()
				return False

			ret = self.download_cetk(
				writer
			)
			assistance.notify(CDNDownloadStage.CETK, None, False, ret, self)
			if assistance.in_error():
				assistance.abort()
				return False

		writer = assistance.notify(CDNDownloadStage.TMD, None, True, 0, self)
		if writer is None and assistance.in_error():
			assistance.abort()
			return False

		ret = self.download_tmd(
			writer
		)
		assistance.notify(CDNDownloadStage.TMD, None, False, ret, self)
		if ret != 200 or assistance.in_error():
			assistance.abort()
			return False

		dlc = Title.is_ctrdlc(self._tid)

		for i in self._tmd.content_chunk_records:
			right = not dlc or \
				not right_check or \
				self._tik is None or \
				self._tik.has_rights_to_index(i.index)
			if not right:
				continue

			writer = assistance.notify(CDNDownloadStage.CONTENT, i, True, 0, self)
			if writer is None:
				if assistance.in_error():
					assistance.abort()
					return False
				continue

			ret = self.download_content_id(
				i.id,
				writer
			)
			assistance.notify(CDNDownloadStage.CONTENT, i, False, ret, self)
			if assistance.in_error():
				assistance.abort()
				return False

			if not content_error_continue and ret != 200:
				assistance.abort()
				return False

		assistance.notify(CDNDownloadStage.END, None, False, 0, self)
		if assistance.in_error():
			assistance.abort()
			return False

		return True

	def add_additional_cert(self, new_cert: Certificate) -> None:
		cert = self._certs.get(new_cert.full_issuer, None)
		if cert is not None and bytes(cert) != bytes(new_cert):
			raise DataProcessingError("Discrepancy found, duplicated cert mismatch previous downloaded one")

		self._certs[new_cert.full_issuer] = new_cert



import json, base64, typing, warnings
from collections import OrderedDict
from struct import pack

from cryptography.hazmat.primitives.asymmetric import ec, utils as crypto_utils
from cryptography.hazmat.primitives import hashes

from .boottimeemu import BootTimeEmu
from .otp import OTP, CTCert
from .apcert import APCert
from .secureinfo import SecureInfo
from .movablesed import MovableSed
from .title import Title, MediaType
from ._sys_titles import o3ds_sys_titles, n3ds_sys_titles
from .exception import ClassInitError, DataProcessingError
from . import regionaldata
from ._py_ver_fixes import CTRModel_T

__all__ = [
	"SimpleCtrDevice"
]

class SimpleCtrDevice:
	"""
	Mostly simple class to simulate basic functionalities and information.
	Specifically:
	 - Boot time
	 - Region, country and language
	 - Serial number
	 - OTP, CTCert, generating and signing with APCerts
	 - Ecommerce
	"""

	def __init__(
		self,
		*,
		json_fp: typing.Optional[typing.TextIO] = None,
		json_file: typing.Optional[str] = None,
		json_string: typing.Optional[str] = None,
		time_padding_emu: bool = True,
		**kwargs
	):
		self._boottime = BootTimeEmu()
		self._time_padding_emu = bool(time_padding_emu) # except boot time, controls padding times

		try:
			if json_fp:
				settings = json.load(json_fp)
			elif json_file:
				with open(json_file, 'r', encoding='utf-8') as file:
					settings = json.load(file)
			else:
				settings = json.loads(json_string)
		except Exception as e:
			raise ClassInitError("Json loading error") from e

		try:
			self._otp = OTP(base64.b64decode(settings['otp'][:0x158]))
			self._ctcert = CTCert(self._otp)
			self._secinfo = SecureInfo(base64.b64decode(settings['secureinfo'][:0x16C]), self._otp.is_dev)
		except Exception as e:
			raise ClassInitError("Failed loading OTP, CTCert or SecureInfo") from e

		self._msed = settings.get('msed', None)
		if self._msed is None:
			self._msed = MovableSed.generate_from_otp(self._otp)
			self._msed.fixup_footer_data(self._otp)
		else:
			self._msed = MovableSed(base64.b64decode(self._msed[:0x1AC]))

		self._model_override = settings.get('model', None)

		self._region = regionaldata.Region.get_region_str(settings.get('region', self._secinfo.region))
		self._country = regionaldata.Country.get_country_str(settings.get('country', None))
		self._language = regionaldata.Language.get_language_str(settings.get('language', None))

		self._orig_region = self._secinfo.region_str

		self._titles = {
			MediaType.NAND: [],
			MediaType.SD: [],
			MediaType.GAME_CARD: None
		}

		def _try_title(x):
			try:
				return Title(*x)
			except (TypeError, ClassInitError):
				return None

		titles = settings.get('titles', self._titles)
		if isinstance(titles, dict):
			nand_titles = titles.get(MediaType.NAND, [])
			sd_titles = titles.get(MediaType.SD, [])
			cart_title = titles.get(MediaType.GAME_CARD, None)
			nand_titles = nand_titles if isinstance(nand_titles, list) else []
			sd_titles = sd_titles if isinstance(sd_titles, list) else []
			cart_title = cart_title if isinstance(cart_title, list) else None

			tmp_nand = {}
			tmp_sd = {}

			for i in nand_titles:
				title = _try_title(i)
				if title is None:
					continue
				_title = tmp_nand.get(int(title), None)
				title = title if _title is None else max(title, _title)
				tmp_nand[int(title)] = title

			for i in sd_titles:
				title = _try_title(i)
				if title is None:
					continue
				_title = tmp_sd.get(int(title), None)
				title = title if _title is None else max(title, _title)
				tmp_sd[int(title)] = title

			for i, y in tmp_nand.items():
				self._titles[MediaType.NAND] += y

			del tmp_nand

			for i, y in tmp_sd.items():
				self._titles[MediaType.SD] += y

			del tmp_sd

			if cart_title:
				self._titles[MediaType.GAME_CARD] = _try_title(cart_title)

		self._is_doing_init = 1

		self.region_change(self._region, self._country, self._language) # will validate things

		del self._is_doing_init

		self._ecommerce_info = settings.get('ecommerce_info', {})
		if not isinstance(self._ecommerce_info, dict):
			raise ClassInitError("ECommerce information invalid")

	def serialize_json(
		self,
		*,
		json_fp: typing.Optional[typing.TextIO] = None,
		json_file: typing.Optional[str] = None,
		**kwargs
	) -> typing.Optional[str]:
		settings = OrderedDict()
		settings['otp'] = base64.b64encode(bytes(self._otp)).decode()
		settings['secureinfo'] = base64.b64encode(bytes(self._secinfo)).decode()
		settings['msed'] = base64.b64encode(bytes(self._msed)).decode()
		settings['model'] = self._model_override
		settings['region'] = self._region
		if self._country:
			settings['country'] = self._country
		settings['language'] = self._language
		settings['ecommerce_info'] = self._ecommerce_info
		settings['titles'] = OrderedDict()
		settings['titles'][MediaType.NAND] = [i.export_tuple for i in self._titles[MediaType.NAND]]
		settings['titles'][MediaType.SD] = [i.export_tuple for i in self._titles[MediaType.SD]]
		settings['titles'][MediaType.GAME_CARD] = self._titles[MediaType.GAME_CARD].export_tuple if self._titles[MediaType.GAME_CARD] else None
		if json_fp:
			json.dump(settings, json_fp, indent=2)
		elif json_file:
			with open(json_file, 'w', encoding='utf-8') as file:
				json.dump(settings, file, indent=2)
		else:
			return json.dumps(settings, indent=2)

	@staticmethod
	def generate_new_json(
		*,
		otp_data: typing.Union[OTP, typing.SupportsBytes, None] = None, # otp bytes data
		otp_fp: typing.Optional[typing.BinaryIO] = None, # or otp read binary io
		otp_file: typing.Optional[str] = None, # otherwise otp file path
		secureinfo_data: typing.Union[OTP, typing.SupportsBytes, None] = None, # secureinfo bytes data
		secureinfo_fp: typing.Optional[typing.BinaryIO] = None, # or secureinfo read binary io
		secureinfo_file: typing.Optional[str] = None, # otherwise secureinfo file path
		serialnumber: typing.Optional[str] = None, # not used if secureinfo given
		msed_data: typing.Union[MovableSed, typing.SupportsBytes, None] = None, # generated unsigned from OTP if msed not given
		msed_fp: typing.Optional[typing.BinaryIO] = None, # generated unsigned from OTP if msed not given
		msed_file: typing.Optional[str] = None, # generated unsigned from OTP if msed not given
		model_override: typing.Optional[CTRModel_T] = None, # model to override serial number based detection
		region: typing.Optional[regionaldata.RegionType] = None, # optional if secureinfo is given instead of just serial
		country: typing.Optional[regionaldata.CountryType] = None, # optional if region has only one country
		language: typing.Optional[regionaldata.LanguageType] = None, # default language if not given
		json_fp: typing.Optional[typing.TextIO] = None, # any file-like textio writable object valid for json.dump
		json_file: typing.Optional[str] = None, # any path-like object valid for open if json_fp was not given
		**kwargs
	) -> typing.Optional[str]:
		otp = None

		if otp_data:
			otp = OTP(otp_data)
		elif otp_fp:
			otp = OTP(otp_fp.read(0x100))
		elif otp_file:
			with open(otp_file, 'rb') as i:
				otp = OTP(i.read(0x100))

		if not otp:
			raise DataProcessingError("No OTP loaded")

		secureinfo = None

		if secureinfo_data:
			secureinfo = SecureInfo(secureinfo_data, otp.is_dev)
		elif secureinfo_fp:
			secureinfo = SecureInfo(secureinfo_fp.read(0x111), otp.is_dev)
		elif secureinfo_file:
			with open(secureinfo_file, 'rb') as i:
				secureinfo = SecureInfo(i.read(0x111), otp.is_dev)
		elif serialnumber:
			if regionaldata.Region.get_region(region) is None:
				raise DataProcessingError("Region required if using a serial number instead of a SecureInfo")
			secureinfo = SecureInfo.create_from_serial_number(serialnumber, region, otp.is_dev)

		if not secureinfo:
			raise DataProcessingError("No SecureInfo loaded")

		if msed_data:
			msed = MovableSed(msed_data)
		elif msed_fp:
			msed = MovableSed(msed_fp.read(0x140))
		elif msed_file:
			with open(msed_file, 'rb') as i:
				msed = MovableSed(i.read(0x140))
		else:
			msed = MovableSed.generate_from_otp(otp)

		if model_override is not None and model_override.upper() not in ['CTR', 'SPR', 'FTR', 'KTR', 'RED', 'JAN']:
			raise DataProcessingError("Invalid model override")

		model_override = model_override.upper() if model_override is not None else None

		region = regionaldata.Region.get_region(secureinfo.region if region is None else region)
		country = regionaldata.Country.get_country(country)
		language = regionaldata.Language.get_language(language)

		if region is None:
			raise DataProcessingError("No valid region")

		if region == regionaldata.Region.TWN and language is not None and language == regionaldata.Language.ZH:
			language = regionaldata.Language.ZH_TRAD

		if region == regionaldata.Region.AUS:
			raise DataProcessingError("AUS is a region that has an assigned value, but not properly usable")

		valid_countries = regionaldata.Country.get_region_country_list(region)

		if country is not None:
			if country not in valid_countries:
				raise DataProcessingError("Country not valid for region. Applicable ISO 3166-1 Alpha-2 for this region: " + (", ".join([i.name for i in valid_countries])))
		else:
			if len(valid_countries) > 1:
				raise DataProcessingError("No country specified on a region with more than one option, cannot load a default. Applicable ISO 3166-1 Alpha-2 for this region: " + (", ".join([i.name for i in valid_countries])))

			warnings.warn("Country not specified, but only one option available on region, loading default")
			country = valid_countries[0]

		del valid_countries

		if language is not None:
			valid_languages = regionaldata.Language.get_region_language_list(region)

			if language not in valid_languages:
				warnings.warn("Dubious language for region, loading default")
				language = valid_languages[0]

			del valid_languages
		else:
			language = regionaldata.Language.get_region_language_list(region)[0]

		settings = OrderedDict()
		settings['otp'] = base64.b64encode(bytes(otp)).decode()
		settings['secureinfo'] = base64.b64encode(bytes(secureinfo)).decode()
		settings['msed'] = base64.b64encode(bytes(msed)).decode()
		settings['model'] = model_override
		settings['region'] = regionaldata.Region.get_region_str(region)
		settings['country'] = regionaldata.Country.get_country_str(country)
		settings['language'] = regionaldata.Language.get_language_str(language)
		# ecommerce info and titles are not needed to be init here
		# after all, if missing, init loads defaults
		if json_fp:
			json.dump(settings, json_fp, indent=2)
		elif json_file:
			with open(json_file, 'w', encoding='utf-8') as file:
				json.dump(settings, file, indent=2)
		else:
			return json.dumps(settings, indent=2)

	def reboot(self) -> None:
		self._boottime.reboot()

	def sleep_with_rng_curve(
		self,
		sleep_min_time: typing.Union[int, float],
		rng_multiplier: typing.Union[int, float] = 1
	) -> None:
		self._boottime.sleep_with_rng_curve(sleep_min_time, rng_multiplier)

	def ap_sign(self, title_id: typing.SupportsInt, data: typing.SupportsBytes) -> typing.Tuple[bytes, bytes]:
		data = bytes(data)
		apcert = APCert(self._ctcert, title_id)
		signature = apcert.private_key().sign(data, ec.ECDSA(hashes.SHA256()))
		signature = crypto_utils.decode_dss_signature(signature)
		return (bytes(apcert), pack("30s30s", signature[0].to_bytes(30, 'big'), signature[1].to_bytes(30, 'big')))

	def export_ivs(self) -> bytes:
		ivs = self._msed.produce_ivs()
		if ivs is None:
			raise DataProcessingError("Could not export IVS successfully")
		return ivs

	def import_ivs(self, ivs: typing.SupportsBytes) -> None:
		msed = self._msed.load_from_ivs(ivs, self._otp.is_dev)
		if msed is None:
			raise DataProcessingError("Could not import IVS successfully")
		msed.fixup_footer_data(self._otp)
		self._msed = msed

	def region_change(
		self,
		region: regionaldata.RegionType,
		country: typing.Optional[regionaldata.CountryType] = None,
		language: typing.Optional[regionaldata.LanguageType] = None
	) -> None:
		region = regionaldata.Region.get_region(region)
		country = regionaldata.Country.get_country(country)
		language = regionaldata.Language.get_language(language)

		if region is None:
			raise DataProcessingError("No valid region")

		if region == regionaldata.Region.TWN and language is not None and language == regionaldata.Language.ZH:
			language = regionaldata.Language.ZH_TRAD

		if region == regionaldata.Region.AUS:
			raise DataProcessingError("AUS is a region that has an assigned value, but not properly usable")

		valid_countries = regionaldata.Country.get_region_country_list(region)

		if country is not None:
			if country not in valid_countries:
				raise DataProcessingError("Country not valid for region. Applicable ISO 3166-1 Alpha-2 for this region: " + (", ".join([i.name for i in valid_countries])))
		else:
			if len(valid_countries) > 1:
				raise DataProcessingError("No country specified on a region with more than one option, cannot load a default. Applicable ISO 3166-1 Alpha-2 for this region: " + (", ".join([i.name for i in valid_countries])))

			warnings.warn("Country not specified, but only one option available on region, loading default")
			country = valid_countries[0]

		del valid_countries

		if language is not None:
			valid_languages = regionaldata.Language.get_region_language_list(region)

			if language not in valid_languages:
				warnings.warn("Dubious language for region, loading default")
				language = valid_languages[0]

			del valid_languages
		else:
			language = regionaldata.Language.get_region_language_list(region)[0]

		curr_region = regionaldata.Region.get_region(self._region)

		sys_titles_curr = (n3ds_sys_titles if self.is_n3ds else o3ds_sys_titles)[curr_region]
		sys_titles_new = (n3ds_sys_titles if self.is_n3ds else o3ds_sys_titles)[region]
		sys_titles_curr = sys_titles_curr if sys_titles_curr else []
		sys_titles_new = sys_titles_new if sys_titles_new else []
		if hasattr(self, '_is_doing_init'):
			nand_titles = [i for i in self._titles[MediaType.NAND] if (not Title.is_ctranysys(int(i)) or int(i) in sys_titles_curr)]
			nand_titles += [i for i in sys_titles_curr if int(i) not in nand_titles]
		else:
			nand_titles = list(sys_titles_new).copy() if sys_titles_new else []
			nand_titles += [i for i in self._titles[MediaType.NAND] if not Title.is_ctranysys(int(i))]
		self._titles[MediaType.NAND] = nand_titles

		self._region = regionaldata.Region.get_region_str(region)
		self._country = regionaldata.Country.get_country_str(country)
		self._language = regionaldata.Language.get_language_str(language)

	def country_change(self, country: typing.Optional[regionaldata.CountryType]) -> None:
		region = regionaldata.Region.get_region(self._region)
		country = regionaldata.Country.get_country(country)

		valid_countries = regionaldata.Country.get_region_country_list(region)

		if country is not None:
			if country not in valid_countries:
				raise DataProcessingError("Country not valid for region. Applicable ISO 3166-1 Alpha-2 for this region: " + (", ".join([i.name for i in valid_countries])))
		else:
			if len(valid_countries) > 1:
				raise DataProcessingError("No country specified on a region with more than one option, cannot load a default. Applicable ISO 3166-1 Alpha-2 for this region: " + (", ".join([i.name for i in valid_countries])))

			warnings.warn("Country not specified, but only one option available on region, loading default")
			country = valid_countries[0]

		del valid_countries

		self._country = regionaldata.Country.get_country_str(country)

	def language_change(self, language: typing.Optional[regionaldata.LanguageType]) -> None:
		region = regionaldata.Region.get_region(self._region)
		language = regionaldata.Language.get_language(language)

		if region == regionaldata.Region.TWN and language is not None and language == regionaldata.Language.ZH:
			language = regionaldata.Language.ZH_TRAD

		if language is not None:
			valid_languages = regionaldata.Language.get_region_language_list(region)

			if language not in valid_languages:
				warnings.warn("Dubious language for region, loading default")
				language = valid_languages[0]

			del valid_languages
		else:
			language = regionaldata.Language.get_region_language_list(region)[0]

		self._language = regionaldata.Language.get_language_str(language)

	def update_ecommerce(self, info: dict) -> None:
		if not isinstance(info, dict):
			raise DataProcessingError("ECommerce information invalid")

		self._ecommerce_info = info.copy()

	def drop_ecommerce(self) -> None:
		self._ecommerce_info = {}

	def titles(self, media: MediaType) -> typing.Iterable[Title]:
		media = MediaType(media)
		if media == MediaType.GAME_CARD:
			return [self._titles[media]] if self._titles[media] else []
		return list(self._titles[media]).copy()

	@property
	def ct_cert(self) -> bytes:
		return bytes(self._ctcert)

	@property
	def ct_cert_full(self) -> CTCert:
		return self._ctcert

	@property
	def msed(self) -> MovableSed:
		return self._msed

	@property
	def boot_time_ms(self) -> int:
		return self._boottime.get_microseconds

	@property
	def ecommerce_info(self) -> dict:
		return self._ecommerce_info.copy()

	@property
	def device_id(self) -> int:
		return self._otp.device_id

	@property
	def serial_no(self) -> str:
		return self._secinfo.serial

	@property
	def model_override(self) -> typing.Optional[CTRModel_T]:
		return self._model_override

	@property
	def original_region(self) -> str:
		"""As per secureinfo"""
		return self._orig_region

	@property
	def region(self) -> str:
		return self._region

	@property
	def country(self) -> str:
		return self._country

	@property
	def language(self) -> str:
		return self._language

	@property
	def is_n3ds(self) -> bool:
		return (self.device_id & 0x80000000) != 0 

	@property
	def is_dev(self) -> bool:
		return self._otp.is_dev

	@property
	def is_retail(self) -> bool:
		return not self._otp.is_dev

	@property
	def has_ecommerce_info(self) -> bool:
		return self._ecommerce_info.get('account_status', None) is not None

	@property
	def time_padding_emu(self):
		return self._time_padding_emu
